#include "Frustum.h"
#include "OpenGL.h"
#include "HomLinAlg.h"
#include "Sphere.h"
#include "Box.h"

	#define rowop(M, R1, OP, R2) Vector<GLfloat>( \
		M(R1,0) OP M(R2,0), \
		M(R1,1) OP M(R2,1), \
		M(R1,2) OP M(R2,2), \
		M(R1,3) OP M(R2,3))

	#define normalize(X)	X /= sqrt( sqr(X(0)) + sqr(X(1)) + sqr(X(2)))


Frustum::Frustum(const Matrix<GLfloat> & MV, const Matrix<GLfloat> & P)
	: 	P 		(P),
		MV		(MV),
		M		(P * MV)
{
	assert(P.nRows() == 4);
	assert(P.nCols() == 4);
	assert(MV.nRows() == 4);
	assert(MV.nCols() == 4);

	for (unsigned int d = 0; d < sizein(planes); 	d++)	// direction
	for (unsigned int n = 0; n < sizein(planes[d]); n++)	// + / -
	{
		if (n == 0) planes[d][n] = rowop(M, 3, +, d);
		else		planes[d][n] = rowop(M, 3, -, d);
		normalize(planes[d][n]);
	}
}

Frustum::~Frustum()
{
}


TopoRel Frustum::getTopoRel(const Sphere & s) const
{
	for (unsigned int d = 0; d < sizein(planes); 	d++)	// direction
	for (unsigned int n = 0; n < sizein(planes[d]); n++)	// + / -
	{
		double l = planes[d][n] * s.x; // signed distance by homogenous inproduct
			 if (l < -s.radius)		return OUTSIDE;
		else if (l < +s.radius) 	return INTERSECT;
	}
	// all tests endured
	return INSIDE;
}



void Frustum::drawTopoRel(const Sphere & s) const
{
	OpenGL::Disable(GL_LIGHTING);

	OpenGL::Begin(GL_LINES);
	OpenGL::Color(1,0,1);

	assert(s.x(3) == 1.0);
	
	for (unsigned int d = 0; d < sizein(planes); 	d++)	// direction
	for (unsigned int n = 0; n < sizein(planes[d]); n++)	// + / -
	{
		double l = planes[d][n] * s.x; // signed distance by homogenous inproduct
		OpenGL::Vertex(s.x);
		OpenGL::Vertex(Vector<GLfloat>(3U, s.x - planes[d][n]*l));
	}
	OpenGL::End();
	OpenGL::Enable(GL_LIGHTING);
	
	
}


/* 	- all points are behind ONE SINGLE plane  => OUT
 	- all points are in front of ALL planes => IN
 	- else intersect
*/

TopoRel Frustum::getTopoRel(const Box 	& b) const
{
	unsigned int fPlane = 0; // box in front of how many planes?

	// iterate over planes	
	for (unsigned int d = 0; d < sizein(planes); 	d++)	// direction
	for (unsigned int n = 0; n < sizein(planes[d]); n++)	// + / -
	{
		unsigned int fPoint = 0; // how many points in front of current plane
		
		// iterate over box vertices
		for (unsigned int bx = 0; bx < 2; 	bx++)	
		for (unsigned int by = 0; by < 2; 	by++)
		for (unsigned int bz = 0; bz < 2; 	bz++)
		{
			// signed distance by homogenous inproduct
			double l = Vector<GLfloat>(b.p[0][bx], b.p[1][by], b.p[2][bz], 1.0) * planes[d][n];
			if (l > 0) fPoint++;
		}

		// all behind means out
		if (fPoint == 0) return OUTSIDE;
        // all in means box in front of a plane
        if (fPoint == 8) fPlane++;
	}

	if (fPlane == 6) 	return INSIDE;
	return INTERSECT;
}



void Frustum::draw() const
{
//	OpenGL::Disable(GL_DEPTH_TEST);
	OpenGL::Disable(GL_LIGHTING);
	OpenGL::DepthMask(GL_FALSE);
//	OpenGL::Enable(GL_BLEND);
	OpenGL::PolygonMode(GL_FRONT_AND_BACK, GL_FILL);
//	OpenGL::BlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	
	OpenGL::PushMatrix();
	OpenGL::MultMatrix(HomMatrix<GLfloat>(M.inv()));

//	OpenGL::Color(1,0,0,0.5);
//	OpenGL::SolidCube(1.999);
	
	OpenGL::Color(1,0,1);
	OpenGL::WireCube(1.999);
	
	OpenGL::PopMatrix();
	
	OpenGL::PolygonMode(GL_FRONT, GL_FILL);
//	OpenGL::Disable(GL_BLEND);
	OpenGL::DepthMask(GL_TRUE);
	OpenGL::Enable(GL_LIGHTING);
//	OpenGL::Enable(GL_DEPTH_TEST);
}

const Matrix<GLfloat> & Frustum::getM() const
{
	return M;
}
