/*------------------------------------------------------------------------------

Wouter Koolen en Steven de Rooij

*/


#ifndef _HOM_LINALG2_H
#define _HOM_LINALG2_H

#include <cassert>
#include "utils.h"
#include "LinAlg.h"

#if !defined __USE_BSD // we need math constants
	#define __USE_BSD
	#include <cmath>
	#undef	__USE_BSD
#else
	#include <cmath>
#endif

#include <iostream>
using namespace std;



#define Base Matrix<T>


template<class T>
class HomMatrix : public Base
{
protected:
    void reHomogenize() 
    {
        if ((*this)(rows-1, cols-1) == 1.0) return;

		T scale = (*this)(rows-1, cols-1);
		
		assert(scale != 0);

        for (unsigned int row = 0; row < rows; row++)
        {
            for (unsigned int col = 0; col < cols; col++)
            {
				set(row, col, (*this)(row,col) / scale);
            }
        }
    }



public:
	HomMatrix() 
	: Base()
	{ /* don't rehomogenize, because this is a fake constructor */ }


	HomMatrix(unsigned int rows, unsigned int cols) 
	: Base(rows, cols)
	{ 
		set(rows-1, cols-1, 1);
		reHomogenize(); 
	}
	

	HomMatrix(unsigned int rows, unsigned int cols, const T x[], bool columnwise = true) 
	: Base(rows, cols, x, columnwise)
	{ reHomogenize(); }


	HomMatrix(const Base & m) : Base(m)
	{ reHomogenize(); }

 	// slow (but never used anyway)
 	HomMatrix scale(T s) const
 	{
 		//operation only exists for square matrices
 		assert(rows == cols);
 		
 		return *this * scaling(s);
 	}
 
 	HomMatrix translate(const Vector<T> & v) const
 	{
		assert(v.rows == rows-1);
 		// operation only exists for square matrices
 		assert(rows == cols);
 		
 		return *this * translation(v) ;
 	}


	/* angle must be double because of limited
		float RANGE (not precision)
	*/
 
 	HomMatrix rotate(double angle, Vector<T> axis) const
 	{
		assert(v.rows == rows-1);
 		// operation only exists for square matrices
 		assert(rows == cols);
 		
 		return *this * rotation(angle, axis);
 	}


	HomMatrix operator * (const HomMatrix & m) const
	{	
		return HomMatrix(Base::operator*(m));
	}

	Vector<T> operator * (const Vector<T> & v) const
	{
		assert(*v == rows-1);
		
		T r[rows];
		
		// homogenize vector
		for (unsigned int i = 0; i < rows-1; i++)
		{
			r[i] = v(i);
		}
		r[rows-1] = 1;
		
		// multiply
		Vector<T> r2 = Base::operator *(Vector<T>(r));
		
		double scale = r2(rows-1);
		
		assert(scale != 0);

		// dishomogenize result vector
		for (unsigned int i = 0; i < rows-1; i++)
		{
			r[i] = r2(i) / scale;
		}

		return Vector<T>(rows-1, r);
	}




	HomMatrix inv() const
	{
		return HomMatrix(Base::inv());
	}


 	static HomMatrix translation(Vector<T> v)
 	{
		unsigned int size = v.dimensionality()+1;
		
		HomMatrix r(size, size);
		
		for (unsigned int row = 0; row < size; row++)
		{
			for (unsigned int col = 0; col < size; col++)
			{
				r.set(row, col, row == col 
								? 1
								: col == size-1
								  ? v(row)
								  : 0);
			}
		}
		
		return r;
 	}
 
 	static HomMatrix scaling(unsigned int size, T s)
 	{
		HomMatrix r(size, size);
		
		for (unsigned int row = 0; row < size; row++)
		{
			for (unsigned int col = 0; col < size; col++)
			{
				r.set(row,col, row == col 
								? col == size-1
								  ? 1
								  : s
								: 0);
			}
		}
		
		return r;
 	}


	/* angle must be double because of limited
		float RANGE (not precision)
	*/
 
 	static HomMatrix rotation(double angle, Vector<T> axis)
 	{
		assert(axis.d() == 3);
		
 		T s = sin(angle);
 		T c = cos(angle);
 		T d = 1-c;
 
 		Vector<T> r = axis.normalize();
		
		T m[] = 
		{
			r(0)*r(0)*d+c, 			r(1)*r(0)*d-r(2)*s,	r(2)*r(0)*d+r(1)*s, 	0,
 			r(0)*r(1)*d+r(2)*s, 	r(1)*r(1)*d+c,		r(2)*r(1)*d-r(0)*s, 	0,
 			r(0)*r(2)*d-r(1)*s, 	r(1)*r(2)*d+r(0)*s,	r(2)*r(2)*d+c, 			0,
 			0,						0,					0,						1,
		};
 
 		return HomMatrix(4, 4, m);
 	}

}; // class HomMatrix


#endif //_HOM_LINALG2_H
