#include "Lamp.h"
#include "OpenGL.h"
#include "utils.h"


// constructor for DIRECTIONAL lights (suns)
Lamp::Lamp(
const Vector<GLfloat> 	& 	ambient,
const Vector<GLfloat> 	& 	diffuse,
const Vector<GLfloat> 	& 	specular,
const Vector<GLfloat> 	& 	dir) 
:
ambient					(ambient), 
diffuse					(diffuse), 
specular				(specular), 
pos						(-dir, 0.0),
lampID(OpenGL::lightbuffer.newLight())
{
	assert(*ambient == 4);
	assert(*diffuse == 4);
	assert(*specular == 4);
	assert(*dir == 3);
	setup();
}
// Constructor ends



// constructor for POSITIONAL light (spots)
Lamp::Lamp(
const Vector<GLfloat> 	& 	ambient,
const Vector<GLfloat> 	& 	diffuse,
const Vector<GLfloat> 	& 	specular,
const Vector<GLfloat> 	& 	pos,
GLfloat						constant_attenuation,
GLfloat						linear_attenuation,
GLfloat						quadratic_attenuation,
GLfloat						spot_exponent,
const Vector<GLfloat>	&	spot_direction,
GLfloat						spot_cutoff) 
:
ambient					(ambient), 
diffuse					(diffuse), 
specular				(specular), 
pos						(pos), 
constant_attenuation	(constant_attenuation),
linear_attenuation		(linear_attenuation),
quadratic_attenuation	(quadratic_attenuation),
spot_exponent			(spot_exponent),
spot_direction			(spot_direction),
spot_cutoff				(spot_cutoff),
lampID(OpenGL::lightbuffer.newLight())
{
	assert(*ambient == 4);
	assert(*diffuse == 4);
	assert(*specular == 4);
	assert(*pos == 4);
	assert(*spot_direction == 3);
	setup();
}
// Constructor ends



Lamp::~Lamp()
{
	OpenGL::Disable(GL_LIGHT0 + lampID);
	OpenGL::lightbuffer.freeLight(lampID);
}


/* Parameters that vary with the transformation, set every frame */
void Lamp::set(bool lit) const
{
	int l = GL_LIGHT0 + lampID;

	if (!lit)
	{
		OpenGL::Disable(l);
	}
	else
	{
		OpenGL::Enable(l);
		// Position (homogenous position can be a direction if w = 0)
		OpenGL::Lightfv(l, GL_POSITION, pos);
		if (pos(3) != 0.0) // so its positional
		{
			OpenGL::Lightfv(l, GL_SPOT_DIRECTION, 	spot_direction);
		}
	}
}


/* All parameters that stay fixed, only set once */

void Lamp::setup() const
{
	int l = GL_LIGHT0 + lampID;

	OpenGL::Enable(l);

	// Colors
	OpenGL::Lightfv(l, GL_AMBIENT,  		ambient);
	OpenGL::Lightfv(l, GL_DIFFUSE,  		diffuse);
	OpenGL::Lightfv(l, GL_SPECULAR,  		specular);

	if (pos(3) != 0.0) // so its positional
	{
		// Attenuation
		OpenGL::Lightf(l,  GL_CONSTANT_ATTENUATION, 	constant_attenuation);
		OpenGL::Lightf(l,  GL_LINEAR_ATTENUATION, 		linear_attenuation);
		OpenGL::Lightf(l,  GL_QUADRATIC_ATTENUATION, 	quadratic_attenuation);

		// Spot
		OpenGL::Lightf(l,  GL_SPOT_EXPONENT, 	spot_exponent);
		OpenGL::Lightf(l,  GL_SPOT_CUTOFF, 		spot_cutoff);
	}
}


void Lamp::draw() const
{
	OpenGL::Disable(GL_LIGHTING);

	OpenGL::Begin(GL_LINES);
	
		if (pos(3) == 0.0)
		{
			// directional light
			OpenGL::Color(1,1,0);
			OpenGL::Vertex(Vector<GLfloat>(0,0,0));
			OpenGL::Vertex(Vector<GLfloat>(pos(0), pos(1), pos(2))*1000);
		}
		else
		{
			// positional light
			OpenGL::Color(0,1,1);
			OpenGL::Vertex(Vector<GLfloat>(pos(0),0,pos(2)));
			OpenGL::Vertex(pos);

			OpenGL::Color(1,0,1);
			OpenGL::Vertex(pos);
			Vector<GLfloat> p(pos(0), pos(1), pos(2));
			OpenGL::Vertex(p + spot_direction * 10);
		}		
	OpenGL::End();

	OpenGL::Enable(GL_LIGHTING);
}

