/*------------------------------------------------------------------------------

Wouter Koolen en Steven de Rooij

*/


#ifndef _LINALG2_H
#define _LINALG2_H

#include <cassert>
#include "utils.h"

#include <iostream>
#include <iomanip>
using namespace std;


template<class T>  // forward declaration for 'friend'
class Matrix;


template<class T>
class Vector
{
	friend class Matrix<T>;
	
protected:
	unsigned int 	rows;
	T 	* 			x;
	

	// internal constructor
	Vector(T x[], unsigned int rows) : rows(rows), x(x)
	{
	}
	

public:
	Vector(unsigned int rows, T zero = 0) : rows(rows), x(new T[rows])
	{
		for (unsigned int i = 0; i < rows; i++)
		{
			x[i] = zero;
		}
	}


	Vector() : rows(0), x(NULL)
	{
	}
	
	// ambiguous, so resolve using spurious parameter
	inline Vector(T x1, bool useless) : rows(1), x(new T[rows])
	{	x[0] = x1;	}
	inline Vector(T x1, T x2): rows(2), x(new T[rows])
	{	x[0] = x1;	x[1] = x2;	}
	inline Vector(T x1, T x2, T x3): rows(3), x(new T[rows])
	{ x[0] = x1; x[1] = x2; x[2] = x3;}
	inline Vector(T x1, T x2, T x3, T x4): rows(4), x(new T[rows])
	{ x[0] = x1; x[1] = x2; x[2] = x3; x[3] = x4; }
	inline Vector(T x1, T x2, T x3, T x4, T x5): rows(5), x(new T[rows])
	{ x[0] = x1; x[1] = x2; x[2] = x3; x[3] = x4; x[4] = x5; }
	inline Vector(T x1, T x2, T x3, T x4, T x5, T x6): rows(6), x(new T[rows])
	{ x[0] = x1; x[1] = x2; x[2] = x3; x[3] = x4; x[4] = x5; x[5] = x6; }
		

	// create from array (copy)
	Vector(unsigned int rows, const T y[]) : rows(rows), x(new T[rows])
	{
		for (unsigned int i = 0; i < rows; i++)
		{
			x[i] = y[i];
		}
	}

	// copy constructor
	Vector(const Vector & v) : rows(v.rows), x(new T[rows])
	{
		assert(v.x != NULL);
		
		for (unsigned int i = 0; i < rows; i++)
		{
			x[i] = v.x[i];
		}
	}
	
	// copy constructor with coefficient sub-selection
	Vector(unsigned int rows, const Vector & v) : rows(rows), x(new T[rows])
	{
		assert(v.x 		!= 		NULL);
		assert(rows 	<= 		v.rows);
		
		for (unsigned int i = 0; i < rows; i++)
		{
			x[i] = v.x[i];
		}
	}
	
	
	// copy constructor with appended coefficient
	Vector(const Vector & v, T xn) : rows(v.rows+1), x(new T[rows])
	{
		assert(v.x != NULL);
		
		for (unsigned int i = 0; i < v.rows; i++)
		{
			x[i] = v.x[i];
		}

		x[v.rows] = xn;
	}

	// copy constructor with appended vector
	Vector(const Vector & v1, const Vector & v2) : rows(v1.rows+v2.rows), x(new T[rows])
	{
		assert(v1.x != NULL);
		assert(v2.x != NULL);

		for (unsigned int i = 0; i < v1.rows; i++)
		{
			x[i] = v1.x[i];
		}

		for (unsigned int i = 0; i < v2.rows; i++)
		{
			x[i+v1.rows] = v2.x[i];
		}		
	}
		
	
	~Vector()
	{
		if(x != NULL) 
		{
			delete [] x;
			x = NULL;
		}
	}

	unsigned int d() const				{ assert(x != NULL); return rows; }
	unsigned int dimensionality() const	{ assert(x != NULL); return rows; }
	unsigned int operator *() const		{ assert(x != NULL); return rows; }


	Vector & operator = (const Vector & v) 
	{
		// allow for reassignment after default constructor
		assert(rows == v.rows || x == NULL);
		
		if (x == NULL)
		{
			rows = 	v.rows;
			x = 	new T[rows];
		}
  
		for (unsigned int i = 0; i < rows; i++)
		{
			x[i] = v.x[i];
		}

		return *this;
	}


	bool operator == (const Vector & v)  const
	{
		// allow for reassignment after default constructor
		assert(x != NULL);
		assert(v.x != NULL);
		assert(rows == v.rows);

		for (unsigned int i = 0; i < rows; i++)
		{
			if (x[i] != v.x[i]) return false;
		}

		return true;
	}

	bool operator != (const Vector & v) const
	{
		return !((*this) == v);
	}					



	// dot/in product
	inline double operator * (const Vector & v) const
	{
		assert(x != NULL); 
		
		double result = 0;

		for (unsigned int i = 0; i < rows; i++)
		{
			result += x[i] * v.x[i];	
		}

		return result;
	}

	// cross product
	Vector operator	^ 	(const Vector & v) const
	{
		assert(x 	!= 	NULL); 
		assert(rows == 	3);

		return Vector(	x[1] * v.x[2] - x[2] * v.x[1],
						x[2] * v.x[0] - x[0] * v.x[2],
						x[0] * v.x[1] - x[1] * v.x[0]);
	}


	// negative
	Vector operator - () const
	{
		assert(x != NULL); 
		
		T * r = new T[rows];
		for (unsigned int i = 0; i < rows; i++)
		{
			r[i] = -x[i];
		}
		return Vector(r, rows);
	}

	// subtraction
	Vector operator	- (const Vector & v) const
	{
		assert(x 	!= 	NULL); 
		assert(v.x 	!= 	NULL); 
		assert(rows == 	v.rows);
		
		T * r = new T[rows];
		for (unsigned int i = 0; i < rows; i++)
		{
			r[i] = x[i] - v.x[i];
		}
		return Vector(r, rows);
	}

	Vector & operator -= (const Vector & v)
	{ 
		assert(x 	!= 	NULL); 
		assert(v.x 	!= 	NULL); 
		assert(rows == 	v.rows);
		
		for (unsigned int i = 0; i < rows; i++)
		{
			x[i] -= v.x[i];
		}
		return (*this);
	}

	// addition
	Vector operator + (const Vector & v) const

	{
		assert(x 	!= 	NULL); 
		assert(v.x 	!= 	NULL); 
		assert(rows == 	v.rows);
		
		T * r = new T[rows];
		for (unsigned int i = 0; i < rows; i++)
		{
			r[i] = x[i] + v.x[i];
		}
		return Vector(r, rows);
	}

	Vector & operator += (const Vector & v)
	{ 
		assert(x 	!= 	NULL); 
		assert(v.x 	!= 	NULL); 
		assert(rows == 	v.rows);
		
		for (unsigned int i = 0; i < rows; i++)
		{
			x[i] += v.x[i];
		}
		return *this;
	}
	

	// scalar multiplication
	Vector operator * (T scale) const
	{
		assert(x 	!= 	NULL); 

		T * r = new T[rows];
		for (unsigned int i = 0; i < rows; i++)
		{
			r[i] = x[i] * scale;
		}
		return Vector(r, rows);
	}

	Vector & operator *= (T scale)
	{ 
		assert(x 	!= 	NULL); 

		for (unsigned int i = 0; i < rows; i++)
		{
			x[i] *= scale;
		}
		return *this;
	}
	

	// scalar division
	Vector operator / (T scale) const
	{
		assert(scale 	!= 		0);
		return (*this) * (1/scale); // one division, not N
	}

	Vector & operator /= (T scale)
	{ 
		assert(scale 	!= 		0);
		return (*this) *= 1/scale;	
	}


	// index
	inline T operator () (unsigned int i) const
	{
		assert(x 	!=	NULL); 
		assert(i 	< 	rows);
		return x[i];
	}


	// cast operator
	operator const T * const  () const
	{
		assert(x 	!= 	NULL); 
		return x;
	}
	

	// cast operator
	template<class X>
	operator Vector<X> () const
	{
		assert(x 	!= 	NULL);
		X * y = new X[rows];
		for (unsigned int i = 0; i < rows; i++)
		{
			y[i] = x[i];
		}
		
		// a bit tricky and useless to make a copy of y here,
		// but the problem is that i can't get template
		// friend functions to work, so we can't acces
		// Vector<X>::Vector(y, rows), the pass-on constructor...
		// so we copy. Don't rely on this function to be fast!
		
		Vector<X> result(rows, y); 
		delete y;
		
		return result;
	}
	
	
	// norm	
	T sqr_length() const
	{
		assert(x 	!= 	NULL); 
		
		T l = 0;
		for (unsigned int i = 0; i < rows; i++) 
		{
			l += sqr(x[i]);
		}

		return l;
	}
	
	// norm square root
	double length() const
	{
		return sqrt((double)sqr_length());
	}
	

	// normalisation
	Vector normalize() const
	{
		double l = length(); 
		assert(l > 0.0);
		return *this / l;
	}


	
}; // class Vector

template<class T>
ostream & operator << (ostream & s, const Vector<T> & v)
{
	s << "(";

	for (unsigned int i = 0; i < *v; i++)
	{
		if (i != 0) s << ", ";
		s << v[i];
	}
	return s << ")";
}





template<class T>
class Matrix
{
protected:
	unsigned int rows;
	unsigned int cols;
	T * 		 x;
	
	inline void set(int row, int col, T v)
	{
		x[row + col * rows] = v;
	}
	

public:

	// default constructor for use in arrays and stuff
	// does not really create a matrix, only a placeholder
	Matrix() : x(NULL)
	{
	}

	// zero matrix
	Matrix(unsigned int rows, unsigned int cols, T zero = 0) : rows(rows), cols(cols), x(new T[rows * cols])
	{
		for (unsigned int row = 0; row < rows; row++)
		{
			for (unsigned int col = 0; col < cols; col++)
			{
				set(row, col, zero);
			}
		}
	}


	// get dimension from array declaration (don't know if this really works yet)
	template<unsigned int rows, unsigned int cols>
	Matrix(const T y[rows][cols]) : rows(rows), cols(cols), x(new T[rows * cols])
	{
		for (unsigned int row = 0; row < rows; row++)
		{
			for (unsigned int col = 0; col < cols; col++)
			{
				set(row, col, y[row][col]);
			}
		}
	}

	// get data from array, row or column major ordening
	Matrix(unsigned int rows, unsigned int cols, const T y[], bool columnwise = true) : rows(rows), cols(cols), x(new T[rows * cols])
	{
		if (columnwise) // shit, transpose necessary
		{
			for (unsigned int row = 0; row < rows; row++)
			{
				for (unsigned int col = 0; col < cols; col++)
				{
					set(row, col, 	y[col + row * cols]);
				}
			}
		}
		else // row-wise, default OpenGL way
		{
			memcpy(x, y, rows*cols*sizeof(T));
		}
	}
	

	// copy constructor	
	Matrix(const Matrix & m) : rows(m.rows), cols(m.cols), x(new T[rows * cols])
	{
		assert (m.x != NULL);
		
		for (unsigned int row = 0; row < rows; row++)
		{
			for (unsigned int col = 0; col < cols; col++)
			{
				set(row, col, m(row, col) );
			}
		}
	}
	
	
	~Matrix()
	{
		if (x != NULL)
		{
			delete [] x;
			x = NULL;
		}
	}
	
	unsigned int nRows() const { return rows; }
	unsigned int nCols() const { return cols; }
	

/*
	// slow (but never used anyway)
	Matrix scale(T s) const
	{
		assert (x != NULL);
		// operation only exists for square matrices
		assert(rows == cols);
		
		return *this * scaling(s);
	}

	Matrix translate(const Vector<T> & v) const
	{
		assert (x != NULL);
		// operation only exists for square matrices
		assert(rows == cols);
		
		return *this * translation(v) ;
	}

	Matrix rotate(T angle, Vector<T> axis) const
	{
		assert (x != NULL);
		// operation only exists for square matrices
		assert(rows == cols);
		
		return *this * rotation(angle, axis);
	}
*/	


	inline T operator () (unsigned int row, unsigned int col) const
	{
		assert (x != NULL);
		assert(row < rows);
		assert(col < cols);
		return x[row + col * rows];
	}

	

	Matrix operator * (const Matrix & m) const
	{
		assert (x 		!= 		NULL);
		assert (m.x 	!= 		NULL);
		assert(cols 	== 		m.rows);
		
		Matrix r(rows, m.cols);

		for (unsigned int rrow = 0; rrow < rows; rrow++)
		{
			for (unsigned int rcol = 0; rcol < m.cols; rcol++)
			{
				T v = 0;
				
				for (unsigned int i = 0; i < cols; i++)
				{
					v += (*this)(rrow, i) * m(i, rcol);
				}
				r.set(rrow, rcol, v);
			}
		}

		return r;
	}

	Vector<T> operator * (const Vector<T> & v) const
	{
		assert (x 		!= 		NULL);
		assert (v.x 	!= 		NULL);
		assert(rows 	== 		v.rows);
		
		T * r = new T[rows];

		for (unsigned int rrow = 0; rrow < rows; rrow++)
		{
			r[rrow] = 0;
			
			for (unsigned int i = 0; i < cols; i++)
			{
				r[rrow] += (*this)(rrow, i) * v(i);
			}
		}

		return Vector<T>(r, rows);
	}
	
	
	
	void operator = (const Matrix<T> & m)
	{
		// allow for reassignment only after default constructor
		assert (rows == m.rows || x == NULL);
		assert (cols == m.cols || x == NULL);
	
		if (x == NULL)
		{
			rows = m.rows;
			cols = m.cols;
			x = new T[rows * cols];
		}

		for (unsigned int row = 0; row < rows; row++)
		{
			for (unsigned int col = 0; col < cols; col++)
			{
				set(row, col, m(row, col));
			}
		}
	}
	


	// cast operator 
	operator const T * () const
	{
		assert (x != NULL);
		return x;
	}


// cast operator
	template<class X>
	operator Matrix<X> () const
	{
		assert(x 	!= 	NULL);
		X * y = new X[rows*cols];
		for (unsigned int i = 0; i < rows*cols; i++)
		{
			y[i] = x[i];
		}

		// a bit tricky and useless to make a copy of y here,
		// but the problem is that i can't get template
		// friend functions to work, so we can't acces
		// Vector<X>::Vector(y, rows), the pass-on constructor...
		// so we copy. Don't rely on this function to be fast!

		Matrix<X> result(rows, cols, y, false);
		delete y;

		return result;
	}	

	
	/* Just dumped the complete inline code for the matrix determinant here
	   This is not clean and not useful for non 4x4 matrices, but it will do */

 	T det() const
 	{
		assert (x != NULL);
		assert(rows == 4 && cols == 4);
     	return
     	  (*this)(0,3) * (*this)(1,2) * (*this)(2,1) * (*this)(3,0) - 
 		  (*this)(0,2) * (*this)(1,3) * (*this)(2,1) * (*this)(3,0) - 
 		  (*this)(0,3) * (*this)(1,1) * (*this)(2,2) * (*this)(3,0) + 
 		  (*this)(0,1) * (*this)(1,3) * (*this)(2,2) * (*this)(3,0) +
 
     	  (*this)(0,2) * (*this)(1,1) * (*this)(2,3) * (*this)(3,0) - 
 		  (*this)(0,1) * (*this)(1,2) * (*this)(2,3) * (*this)(3,0) - 
 		  (*this)(0,3) * (*this)(1,2) * (*this)(2,0) * (*this)(3,1) + 
 		  (*this)(0,2) * (*this)(1,3) * (*this)(2,0) * (*this)(3,1) +
 
     	  (*this)(0,3) * (*this)(1,0) * (*this)(2,2) * (*this)(3,1) - 
 		  (*this)(0,0) * (*this)(1,3) * (*this)(2,2) * (*this)(3,1) - 
 		  (*this)(0,2) * (*this)(1,0) * (*this)(2,3) * (*this)(3,1) + 
 		  (*this)(0,0) * (*this)(1,2) * (*this)(2,3) * (*this)(3,1) +
 
     	  (*this)(0,3) * (*this)(1,1) * (*this)(2,0) * (*this)(3,2) - 
 		  (*this)(0,1) * (*this)(1,3) * (*this)(2,0) * (*this)(3,2) - 
 		  (*this)(0,3) * (*this)(1,0) * (*this)(2,1) * (*this)(3,2) + 
 		  (*this)(0,0) * (*this)(1,3) * (*this)(2,1) * (*this)(3,2) +
 
     	  (*this)(0,1) * (*this)(1,0) * (*this)(2,3) * (*this)(3,2) - 
 		  (*this)(0,0) * (*this)(1,1) * (*this)(2,3) * (*this)(3,2) - 
 		  (*this)(0,2) * (*this)(1,1) * (*this)(2,0) * (*this)(3,3) + 
 		  (*this)(0,1) * (*this)(1,2) * (*this)(2,0) * (*this)(3,3) +
 
     	  (*this)(0,2) * (*this)(1,0) * (*this)(2,1) * (*this)(3,3) - 
 		  (*this)(0,0) * (*this)(1,2) * (*this)(2,1) * (*this)(3,3) - 
 		  (*this)(0,1) * (*this)(1,0) * (*this)(2,2) * (*this)(3,3) + 
 		  (*this)(0,0) * (*this)(1,1) * (*this)(2,2) * (*this)(3,3) ;
 	}
	
	/* Just dumped the complete inline code for the matrix inverse here
	   This is not clean and not useful for non 4x4 matrices, but it will do */
	
	
 	Matrix<T> inv() const
 	{
		assert (x != NULL);
		assert(rows == 4 && cols == 4);
		T r[4*4] = 
		{
     	 (*this)(1,2)*(*this)(2,3)*(*this)(3,1) - (*this)(1,3)*(*this)(2,2)*(*this)(3,1) + (*this)(1,3)*(*this)(2,1)*(*this)(3,2) - 
 		 (*this)(1,1)*(*this)(2,3)*(*this)(3,2) - (*this)(1,2)*(*this)(2,1)*(*this)(3,3) + (*this)(1,1)*(*this)(2,2)*(*this)(3,3),
     	 (*this)(0,3)*(*this)(2,2)*(*this)(3,1) - (*this)(0,2)*(*this)(2,3)*(*this)(3,1) - (*this)(0,3)*(*this)(2,1)*(*this)(3,2) +
 		 (*this)(0,1)*(*this)(2,3)*(*this)(3,2) + (*this)(0,2)*(*this)(2,1)*(*this)(3,3) - (*this)(0,1)*(*this)(2,2)*(*this)(3,3),
     	 (*this)(0,2)*(*this)(1,3)*(*this)(3,1) - (*this)(0,3)*(*this)(1,2)*(*this)(3,1) + (*this)(0,3)*(*this)(1,1)*(*this)(3,2) - 
 		 (*this)(0,1)*(*this)(1,3)*(*this)(3,2) - (*this)(0,2)*(*this)(1,1)*(*this)(3,3) + (*this)(0,1)*(*this)(1,2)*(*this)(3,3),
     	 (*this)(0,3)*(*this)(1,2)*(*this)(2,1) - (*this)(0,2)*(*this)(1,3)*(*this)(2,1) - (*this)(0,3)*(*this)(1,1)*(*this)(2,2) + 
 		 (*this)(0,1)*(*this)(1,3)*(*this)(2,2) + (*this)(0,2)*(*this)(1,1)*(*this)(2,3) - (*this)(0,1)*(*this)(1,2)*(*this)(2,3),
     	 (*this)(1,3)*(*this)(2,2)*(*this)(3,0) - (*this)(1,2)*(*this)(2,3)*(*this)(3,0) - (*this)(1,3)*(*this)(2,0)*(*this)(3,2) + 
 		 (*this)(1,0)*(*this)(2,3)*(*this)(3,2) + (*this)(1,2)*(*this)(2,0)*(*this)(3,3) - (*this)(1,0)*(*this)(2,2)*(*this)(3,3),
     	 (*this)(0,2)*(*this)(2,3)*(*this)(3,0) - (*this)(0,3)*(*this)(2,2)*(*this)(3,0) + (*this)(0,3)*(*this)(2,0)*(*this)(3,2) - 
 		 (*this)(0,0)*(*this)(2,3)*(*this)(3,2) - (*this)(0,2)*(*this)(2,0)*(*this)(3,3) + (*this)(0,0)*(*this)(2,2)*(*this)(3,3),
     	 (*this)(0,3)*(*this)(1,2)*(*this)(3,0) - (*this)(0,2)*(*this)(1,3)*(*this)(3,0) - (*this)(0,3)*(*this)(1,0)*(*this)(3,2) + 
 		 (*this)(0,0)*(*this)(1,3)*(*this)(3,2) + (*this)(0,2)*(*this)(1,0)*(*this)(3,3) - (*this)(0,0)*(*this)(1,2)*(*this)(3,3),
     	 (*this)(0,2)*(*this)(1,3)*(*this)(2,0) - (*this)(0,3)*(*this)(1,2)*(*this)(2,0) + (*this)(0,3)*(*this)(1,0)*(*this)(2,2) - 
 		 (*this)(0,0)*(*this)(1,3)*(*this)(2,2) - (*this)(0,2)*(*this)(1,0)*(*this)(2,3) + (*this)(0,0)*(*this)(1,2)*(*this)(2,3),
     	 (*this)(1,1)*(*this)(2,3)*(*this)(3,0) - (*this)(1,3)*(*this)(2,1)*(*this)(3,0) + (*this)(1,3)*(*this)(2,0)*(*this)(3,1) - 
 		 (*this)(1,0)*(*this)(2,3)*(*this)(3,1) - (*this)(1,1)*(*this)(2,0)*(*this)(3,3) + (*this)(1,0)*(*this)(2,1)*(*this)(3,3),
     	 (*this)(0,3)*(*this)(2,1)*(*this)(3,0) - (*this)(0,1)*(*this)(2,3)*(*this)(3,0) - (*this)(0,3)*(*this)(2,0)*(*this)(3,1) + 
 		 (*this)(0,0)*(*this)(2,3)*(*this)(3,1) + (*this)(0,1)*(*this)(2,0)*(*this)(3,3) - (*this)(0,0)*(*this)(2,1)*(*this)(3,3),
     	 (*this)(0,1)*(*this)(1,3)*(*this)(3,0) - (*this)(0,3)*(*this)(1,1)*(*this)(3,0) + (*this)(0,3)*(*this)(1,0)*(*this)(3,1) -
 		 (*this)(0,0)*(*this)(1,3)*(*this)(3,1) - (*this)(0,1)*(*this)(1,0)*(*this)(3,3) + (*this)(0,0)*(*this)(1,1)*(*this)(3,3),
     	 (*this)(0,3)*(*this)(1,1)*(*this)(2,0) - (*this)(0,1)*(*this)(1,3)*(*this)(2,0) - (*this)(0,3)*(*this)(1,0)*(*this)(2,1) + 
 		 (*this)(0,0)*(*this)(1,3)*(*this)(2,1) + (*this)(0,1)*(*this)(1,0)*(*this)(2,3) - (*this)(0,0)*(*this)(1,1)*(*this)(2,3),
     	 (*this)(1,2)*(*this)(2,1)*(*this)(3,0) - (*this)(1,1)*(*this)(2,2)*(*this)(3,0) - (*this)(1,2)*(*this)(2,0)*(*this)(3,1) + 
 		 (*this)(1,0)*(*this)(2,2)*(*this)(3,1) + (*this)(1,1)*(*this)(2,0)*(*this)(3,2) - (*this)(1,0)*(*this)(2,1)*(*this)(3,2),
     	 (*this)(0,1)*(*this)(2,2)*(*this)(3,0) - (*this)(0,2)*(*this)(2,1)*(*this)(3,0) + (*this)(0,2)*(*this)(2,0)*(*this)(3,1) - 
 		 (*this)(0,0)*(*this)(2,2)*(*this)(3,1) - (*this)(0,1)*(*this)(2,0)*(*this)(3,2) + (*this)(0,0)*(*this)(2,1)*(*this)(3,2),
     	 (*this)(0,2)*(*this)(1,1)*(*this)(3,0) - (*this)(0,1)*(*this)(1,2)*(*this)(3,0) - (*this)(0,2)*(*this)(1,0)*(*this)(3,1) + 
 		 (*this)(0,0)*(*this)(1,2)*(*this)(3,1) + (*this)(0,1)*(*this)(1,0)*(*this)(3,2) - (*this)(0,0)*(*this)(1,1)*(*this)(3,2),
     	 (*this)(0,1)*(*this)(1,2)*(*this)(2,0) - (*this)(0,2)*(*this)(1,1)*(*this)(2,0) + (*this)(0,2)*(*this)(1,0)*(*this)(2,1) - 
 		 (*this)(0,0)*(*this)(1,2)*(*this)(2,1) - (*this)(0,1)*(*this)(1,0)*(*this)(2,2) + (*this)(0,0)*(*this)(1,1)*(*this)(2,2),
		};

     	return Matrix(4,4,r);
 	}
	
	
	static Matrix identity(unsigned int size)
	{             
		Matrix I(size, size);
		for (unsigned i = 0; i < size; i++)
		{
			I.set(i,i,1);
		}
		return I;
	}

	template <class X>
	friend ostream & operator << (ostream & s, const Matrix<X> & m);

}; // class Matrix

template<class T>
ostream & operator << (ostream & s, const Matrix<T> & m)
{
	for (unsigned int row = 0; row < m.rows; row++)
	{
		if (row != 0) s << endl;
		s << "| ";
		for (unsigned int col = 0; col < m.cols; col++)
		{
			s << setw(10) << m(row, col);
		}
		s << "|";
	}
	return s;
}




#endif //_LINALG2_H
