#include "GL/HomLinAlg.h"
#include "GL/Material.h"
#include "GL/Frustum.h"
#include "GL/Box.h"
#include "GL/Sphere.h"
#include "GL/GLModelPart.h"
#include "GL/GLDisplayList.h"
#include "GL/Frustum.h"
#include "GL/SP.h"

#include "GridModel.h"
#include "utils.h"
#include "array.h"


#include <time.h>
using namespace std;

// #define HEAVY_DUTY_CHECKS 1


Material GridModel::GridMaterial = Material(
 Vector<GLfloat>(	0.1,	0.1, 	0.1, 	0),
 Vector<GLfloat>(	0.3,	0.6, 	0.3,	0),
 Vector<GLfloat>(	0.0,  	0.8, 	0.0, 	0),
 Vector<GLfloat>(	0, 		0, 		0,		0),
 40);



const int GridModel::ns[2][3][2] = 
{
	{{0,0}, {1, 1}, {1, 0}},
	{{0,0}, {0, 1}, {1, 1}},
};



const int GridModel::fn[6][4] =
{
	{ 0, -1, 1, 2},
	{-1, -1, 0, 1},
	{-1, -1, 1, 1},
	{-1,  0, 0, 2},
	{ 0,  0, 1, 1},
	{ 0,  0, 0, 1},
};

class GridModel;


/* 	A grid patch is a square piece of the grid.
	Its location is (x0, y0) - (x1, y1) (END-INCLUSIVE!!!)
	Multiple patches can thus share their borders!
*/

class GridPatch : public SPLeaf, public GLModelPart
{
private:
	unsigned int 		x0, x1, sx;
	unsigned int 		y0, y1, sy;
	double				zmin, zmax;
	GridModel & 		m;
	GLuint ** 			indexbuffer;
	GLDisplayList		dl;
	
public:

	GridPatch(unsigned int x0, unsigned int x1, unsigned int y0, unsigned int y1, GridModel & m)
		: x0(x0), x1(x1), sx(x1 - x0), y0(y0), y1(y1), sy(y1 - y0), m(m)
	{
		assert(x0 < x1);
		assert(y0 < y1);
		
        indexbuffer = new_array<GLuint>(sx, 2*sy+2);
        
		for (unsigned int x = x0; x < x1; x++)
		{
			GLuint * pos = indexbuffer[x-x0];

			for (unsigned int y = y0; y <= y1; y++)
			{
				for (unsigned int u = 0; u < 2; u++)
				{
					*pos++ = x+1-u + y*m.s;
				}
			}
		}
		invalidate();
	}

	virtual ~GridPatch()
	{
		delete_array<GLuint>(indexbuffer, sx, 2*sy+2);
	}

	void draw() //const
	{
		dl.draw(*this);
	}

	void fillDL() const
	{
		for (unsigned int x = 0; x < sx; x++)
		{
			OpenGL::DrawElements(GL_TRIANGLE_STRIP, sy*2+2, GL_UNSIGNED_INT, indexbuffer[x]);
		}
	}
	
	void invalidate()
	{
		SPLeaf::invalidate();
		dl.invalidate();
	}

	Box * recalcBound()
	{
		// research minimum and maximum
		zmin= zmax = m.h[y0][x0];
		for (unsigned int y = y0; y <= y1; y++)
		for (unsigned int x = x0; x <= x1; x++)
		{
				 if (m.h[y][x] < zmin) zmin = m.h[y][x];
			else if (m.h[y][x] > zmax) zmax = m.h[y][x];
		}
		return new Box(x0, x1, zmin, zmax, y0, y1);
	}

	friend class GridModel;
};


#define E_HEIGHT 0
#define APPROX	40


GridModel::GridModel(int logsize) : s(1 << logsize)
{
	h = 		new_array< double >(s, s);
	normals = 	new_array< Vector<GLfloat> >(s, s, 2);				
	vns = 		new_array< Vector<GLfloat> >(s, s);

	// make a nice landscape
	TIME( generate() )
	
	// smooth just a little
	TIME( smooth(0,s,0,s) )
	

	// compute face normals
	TIME( computeFaceNormals(0,s,0,s) )
	
	// compute vertex normals
	TIME( computeVertexNormals(0,s,0,s) )
	

	// make strips of vertices and normals
	vertexbuffer = new GLfloat[3 * s * s];
	normalbuffer = new GLfloat[3 * s * s];
	TIME( bufferize(0,s,0,s) )

	// divide the grid in patches
	patchesx = w_max(1, s/32);
	patchesy = w_max(1, s/32);

	patches = 	new_array<GridPatch*>		(patchesx, patchesy);
	TIME( patch() )

	sp = new SPNode();
	TIME( partition(sp, 0, 0, patchesx, patchesy) )
}


GridModel::~GridModel()
{
	delete sp;
	delete [] vertexbuffer;
	delete [] normalbuffer;
	delete_array(h, 			s, 		s);
	delete_array(vns, 			s, 		s);
	delete_array(normals, 		s, 		s, 	2);
	delete_array(patches,		patchesx, patchesy);
}

void GridModel::patch()
{
	for (unsigned int xi = 0; xi < patchesx; xi++)
	for (unsigned int yi = 0; yi < patchesy; yi++)
	{
		patches[xi][yi] =	new GridPatch(
			xi*(s/patchesx), (xi+1)*(s/patchesx) - (xi == patchesx-1 ? 1 : 0),
			yi*(s/patchesy), (yi+1)*(s/patchesy) - (yi == patchesy-1 ? 1 : 0), *this
		);
	}
}

void GridModel::partition(SPNode * sp, unsigned int x0, unsigned int y0, unsigned int sx, unsigned int sy)
{
	if (sx == 1)
	{
		assert(sy == 1);
		sp->addChild(patches[x0][y0]);
	}
	else
	{
		for (unsigned int xi = 0; xi < 2; xi++)
		for (unsigned int yi = 0; yi < 2; yi++)
		{
			SPNode * child = new SPNode();
			sp->addChild(child);
			partition(child, x0+xi*sx/2, y0+yi*sy/2, sx/2, sy/2);
		}
	}
}





void GridModel::drawVertexNormals(double length) const
{
	OpenGL::Disable(GL_LIGHTING);

	OpenGL::Color(1, 0, 1);
	OpenGL::Begin(GL_LINES);

	// for every vertex
	for (unsigned int x = 0; x < s; x++)
	{
		for (unsigned int y = 0; y < s; y++)
		{
			Vector<GLfloat> v(x, get(h, x, y), y);
			OpenGL::Vertex(v);
			OpenGL::Vertex(v + get(vns, x, y) * length);
		}
	}
	OpenGL::End();				
	OpenGL::Enable(GL_LIGHTING);					
}


void GridModel::drawFaceNormals(double length) const
{
	OpenGL::Disable(GL_LIGHTING);

	OpenGL::Color(1, 1, 0);
	OpenGL::Begin(GL_LINES);		

	// for every triangle
	for (unsigned int x = 0; x < s; x++)
	{
		for (unsigned int y = 0; y < s; y++)
		{
			for (unsigned int k = 0; k < sizein(ns); k++)
			{
				// Calculate the middle
				Vector<GLfloat> v(3);

				for (unsigned int l = 0; l < sizein(ns[k]); l++)
				{
					int x2 = x + ns[k][l][0];
					int y2 = y + ns[k][l][1];

					v = v + Vector<GLfloat>(x2, get(h, x2, y2), y2) 
							/ sizein(ns[k]);
				}

				OpenGL::Vertex(v);
				OpenGL::Vertex(v + get(normals, x, y)[k] * length);
			}
		}
	}
	OpenGL::End();				
	OpenGL::Enable(GL_LIGHTING);		
}

double GridModel::getH(const Vector<GLfloat> & xz) const
{

	double 	x = 	xz(0)+(double)s/2;   
	double 	z = 	xz(1)+(double)s/2;	

	int 	ix = 	(int)x;
	int 	iz =	(int)z;

	double	dxa =	(x-ix); double dxb = 1-dxa;
	double 	dza =	(z-iz); double dzb = 1-dza;
//	double 	dx = 	1-(x - ix);
//	double 	dz = 	1-(z - (int)z);
	
	if (ix < 0 || ix+1 >= (int)s ||
		iz < 0 || iz+1 >= (int)s) return 0.0;

	// bi-linear interpolation
	return dxb * dzb *	h[iz]  [ix] 	+
		   dxa * dzb *	h[iz]  [ix+1]	+
		   dxb * dza * 	h[iz+1][ix]	+
		   dxa * dza *	h[iz+1][ix+1]	;
}


void GridModel::draw(const Frustum & f, bool drawBounds)  const
{
	double  d = s/2.0;
	
	GridMaterial.set();
	
	OpenGL::PushMatrix();
	HomMatrix<GLfloat> transform = HomMatrix<GLfloat>::translation(Vector<GLfloat>(-d,0,-d));
	OpenGL::MultMatrix(transform);
	Frustum f2(transform, f.getM());
	
	OpenGL::VertexPointer(3, GL_FLOAT, 0, vertexbuffer);
	OpenGL::NormalPointer(	 GL_FLOAT, 0, normalbuffer);
	
	OpenGL::EnableClientState(GL_VERTEX_ARRAY);
	OpenGL::EnableClientState(GL_NORMAL_ARRAY);

	sp->check(f2, drawBounds);

	OpenGL::DisableClientState(GL_VERTEX_ARRAY);
	OpenGL::DisableClientState(GL_NORMAL_ARRAY);
	
//	drawVertexNormals(1.0);
//	drawFaceNormals(1.0);
	
	OpenGL::PopMatrix();

}




void GridModel::generate()
{
	randomize();

	h[0][0] = (double)rand()/RAND_MAX*E_HEIGHT;
  
	for (unsigned int i = s; i !=0; i >>= 1)
	{
		unsigned int d = i/2;
		// middenpuntjes
		for (unsigned int x = d; x < s; x += i )
		{
			for (unsigned int y = d; y < s; y += i )
			{
				put(h, x,y,
			 		( get(h, x-d, y-d)+ 
					  get(h, x+d, y+d)+
					  get(h, x-d, y+d)+
					  get(h, x+d, y-d)
					)/4
					+ mod(i*M_SQRT2));
			}
		}

		//randpuntjes
		for (unsigned int t1 = 0; t1 < s; t1 += i )
		{
			for (unsigned int t2 = d; t2 < s; t2 += i )
			{
				// vierkantjes, x=t1, y=t2
				put(h, t1, t2,
					( get(h, t1, t2-d) + 
					  get(h, t1, t2+d) +
					  get(h, t1-d, t2) +
					  get(h, t1+d, t2)
					)/4
					+ mod(i));

				// rondjes, x=t2, y = t1
				put(h, t2, t1,
					( get(h, t2-d, t1) + 
			 		  get(h, t2+d, t1) +
					  get(h, t2, t1-d) +
			 		  get(h, t2, t1+d)
			    	)/4
					+ mod(i));
			}
		}
	}
}

void GridModel::smooth(unsigned int x0, unsigned int x1, unsigned int y0, unsigned int y1)
{
	assert (x0 < x1);
	assert (y0 < y1);

	float kernel[3][3] = {  {1.0/9.0, 1.0/9.0, 1.0/9.0},
							{1.0/9.0, 1.0/9.0, 1.0/9.0},
							{1.0/9.0, 1.0/9.0, 1.0/9.0}};
							
	unsigned int sy = y1-y0;
	unsigned int sx = x1-x0;
	
	unsigned int ksy = sizein(kernel);
	unsigned int ksx = sizein(kernel[0]);

	assert(ksy % 2 == 1);	// kernel must have odd size
	assert(ksx % 2 == 1);	
	

	// create a working copy of the old values
	// we need (skx-1)/2 stretch on both sides
	double ** oldh = 	new_array<double>(sy+ksy-1, sx+ksx-1);

	// fill in working copy
	for (int oy = y0-(int)ksy/2; oy < (int)y1+(int)ksy/2; oy++)
	for (int ox = x0-(int)ksx/2; ox < (int)x1+(int)ksx/2; ox++)
	{
		// watch out for borders of h!
		// fill in gaps with nearest known point
		int tx = ox-(int)x0+ksy/2;
		int ty = oy-(int)y0+ksy/2;
		int fx = w_clamp(0, ox, (int)s-1);
		int fy = w_clamp(0, oy, (int)s-1);
		oldh[ty][tx] = h[fy][fx];
	}
		
		
	// smooth the designated area
	for (unsigned int hy = y0; hy < y1; hy++)
	for (unsigned int hx = x0; hx < x1; hx++)
	{
		h[hy][hx] = 0.0;

		for (unsigned int ky = 0; ky < ksy;	ky++)
		for (unsigned int kx = 0; kx < ksx;	kx++)
		{
			h[hy][hx] 	+= 	kernel[ksy-1-ky][ksx-1-kx]
						*	oldh[hy-y0+ky][hx-x0+kx];
/*			int ny = (int)hy + (int)sky/2 - (int)ky;
			int nx = (int)hx + (int)skx/2 - (int)kx;
			h[hy][hx] += kernel[ky][kx] * 	((nx >= 0 && nx < (int)s && 
											 ny >= 0 && ny < (int)s)
											? oldh[ny][nx]
											: oldh[hy][hx]);
*/											
		}
	}
	delete_array(oldh, sy+ksy-1, sx+ksx-1);
}


unsigned int GridModel::improve(int x) const
{
	int rx = x%s; 
	if (rx < 0) rx += s;
	return rx;
}

template<class T>
T GridModel::get(T ** h, int x, int y) const
{
	return h[improve(y)][improve(x)];
}

template<class T>
void GridModel::put(T ** h, int x, int y, T v)
{
	h[improve(y)][improve(x)] = v;
}







double GridModel::mod(double s)
{
	// We use the Box-Muller transformation to obtain a normally
	// distributed value with mu=0 and sigma=1
	double u1 = (double)rand()/RAND_MAX;
	double u2 = (double)rand()/RAND_MAX;
	double n1 = sqrt(-2*log(u1))*cos(2*M_PI*u2);

	// Return the number with correctly scaled sigma
	const double scale = 0.14;

	return s*n1*scale;
}





void GridModel::computeFaceNormals(unsigned int x0, unsigned int x1, unsigned int y0, unsigned int y1)
{
	assert(x0 < x1);
	assert(y0 < y1);
	assert(x1 <= s);
	assert(y1 <= s);
	
	// Iterate over all triangles
	for (unsigned int j = y0; j < y1; j++)
	{
		for (unsigned int i = x0; i < x1; i++)
		{
			for (unsigned int k = 0; k < sizein(ns); k++)
			{
				Vector<GLfloat> v[3];

				for (unsigned int l = 0; l < sizein(ns[k]); l++)
				{
					int 	x = i + ns[k][l][0];
					int 	y = j + ns[k][l][1];
					double 	z 	= (x >= 0 && x < (int)s && 
								   y >= 0 && y < (int)s)
								? h[y][x]
								: 0;
					v[l] = Vector<GLfloat>(x, z, y);
				}

				Vector<GLfloat> e1 = v[1] - v[0];
				Vector<GLfloat> e2 = v[2] - v[1];
				
				normals[j][i][k] = (e1 ^ e2).normalize();
			}
/*

				Vector<GLfloat> e11	= Vector<GLfloat>(
					ns[0][1][0]	- ns[0][0][0],
					((i + ns[0][1][0] >= 0 && i + ns[0][1][0] < (int)s && j + ns[0][1][1] >= 0 && j + ns[0][1][1] < (int)s) ? h[j + ns[0][1][1]][i + ns[0][1][0]] : 0) - ((i + ns[0][0][0] >= 0 && i + ns[0][0][0] < (int)s && j + ns[0][0][1] >= 0 && j + ns[0][0][1] < (int)s) ? h[j + ns[0][0][1]][i + ns[0][0][0]] : 0),
					ns[0][1][1] - ns[0][0][1]
				);

				Vector<GLfloat> e21	= Vector<GLfloat>(
					ns[0][2][0] - ns[0][1][0], 
					((i + ns[0][2][0] >= 0 && i + ns[0][2][0] < (int)s && j + ns[0][2][1] >= 0 && j + ns[0][2][1] < (int)s) ? h[j + ns[0][2][1]][i + ns[0][2][0]] : 0) - ((i + ns[0][1][0] >= 0 && i + ns[0][1][0] < (int)s && j + ns[0][1][1] >= 0 && j + ns[0][1][1] < (int)s) ? h[j + ns[0][1][1]][i + ns[0][1][0]] : 0),
					ns[0][2][1] - ns[0][1][1]
				);
				
				Vector<GLfloat> e12	= Vector<GLfloat>(ns[1][1][0], (i + ns[1][1][0] >= 0 && i + ns[1][1][0] < (int)s && j + ns[1][1][1] >= 0 && j + ns[1][1][1] < (int)s) ? h[j + ns[1][1][1]][i + ns[1][1][0]] : 0, ns[1][1][1])
										- Vector<GLfloat>(ns[1][0][0], (i + ns[1][0][0] >= 0 && i + ns[1][0][0] < (int)s && j + ns[1][0][1] >= 0 && j + ns[1][0][1] < (int)s) ? h[j + ns[1][0][1]][i + ns[1][0][0]] : 0, ns[1][0][1]);
				Vector<GLfloat> e22	= Vector<GLfloat>(ns[1][2][0], (i + ns[1][2][0] >= 0 && i + ns[1][2][0] < (int)s && j + ns[1][2][1] >= 0 && j + ns[1][2][1] < (int)s) ? h[j + ns[1][2][1]][i + ns[1][2][0]] : 0, ns[1][2][1]) 
										- Vector<GLfloat>(ns[1][1][0], (i + ns[1][1][0] >= 0 && i + ns[1][1][0] < (int)s && j + ns[1][1][1] >= 0 && j + ns[1][1][1] < (int)s) ? h[j + ns[1][1][1]][i + ns[1][1][0]] : 0, ns[1][1][1]);
										
				normals[j][i][0] = (e11 ^ e21).normalize();
				normals[j][i][1] = (e12 ^ e22).normalize();
*/				
		}
	}
	
#ifdef HEAVY_DUTY_CHECKS
// keep this in sync with above code!
// Iterate over all triangles
	for (unsigned int j = 0; j < s; j++)
	{
		for (unsigned int i = 0; i < s; i++)
		{
			for (unsigned int k = 0; k < sizein(ns); k++)
			{
				Vector<GLfloat> v[3];
				for (unsigned int l = 0; l < sizein(ns[k]); l++)
				{
					int 	x = i + ns[k][l][0];
					int 	y = j + ns[k][l][1];
					double 	z = (x >= 0 && x < (int)s && y >= 0 && y < (int)s)
								? get(h, x,y) : 0;
					v[l] = Vector<GLfloat>(x, z, y);
				}

				Vector<GLfloat> e1 = v[1] - v[0];
				Vector<GLfloat> e2 = v[2] - v[1];

				assert(get(normals, i, j)[k] == (e1 ^ e2).normalize());
			}
		}
	}
#endif // HEAVY_DUTY_CHECKS		
}
	

void GridModel::computeVertexNormals(unsigned int x0, unsigned int x1, unsigned int y0, unsigned int y1)
{
	// Iterate over all vertices
	for (unsigned int j = y0; j < y1; j++)
	{
		for (unsigned int i = x0; i < x1; i++)
		{
			Vector<GLfloat> sum(3);

			for (unsigned int k = 0; k < sizein(fn); k++)
			{
				int x = i + fn[k][0];
				int y = j + fn[k][1];
				Vector<GLfloat> n 	= ( x >= 0 && x < (int)s && 
										y >= 0 && y < (int)s)
									? normals[y][x][fn[k][2]] 
									: Vector<GLfloat>(0, 1, 0);
				sum += sum + n * fn[k][3];
			}
			vns[j][i] = sum.normalize();
		}
	}

#ifdef HEAVY_DUTY_CHECKS
// keep this in sync with above code!

	// Iterate over all vertices
	for (unsigned int j = 0; j < s; j++)
	{
		for (unsigned int i = 0; i < s; i++)
		{
			Vector<GLfloat> sum(3);

			for (unsigned int k = 0; k < sizein(fn); k++)
			{
				int x = i + fn[k][0];
				int y = j + fn[k][1];
				Vector<GLfloat> n 	= (x >= 0 && x < (int)s && y >= 0 && y < (int)s)
									? get(normals, x, y)[fn[k][2]]
									: Vector<GLfloat>(0, 1, 0);
				sum += sum + n * fn[k][3];

			}
			for (unsigned int k = 0; k < *get(vns,i,j); k++)
			{
				if( abs(get(vns, i, j)(k) - sum.normalize()(k)) < 0.0001)
				{
				}
				else
				{
					cout<< "outlier: (" << i << ", " << j << ")  is "
						<< get(vns, i, j) << " ? " << sum.normalize()  << "  in rect "
						<< "(" << x0 << ", " << y0 << ") - "
						<< "(" << x1 << ", " << y1 << ")" << " but is not updated" << endl;
					assert(false);					
				}
			}
			
		}
	}
#endif //  HEAVY_DUTY_CHECKS	
}



void GridModel::bufferize(unsigned int x0, unsigned int x1, unsigned int y0, unsigned int y1)
{
	assert(x0 < x1);
	assert(y0 < y1);
	assert(x1 <= s);
	assert(y1 <= s);
	
	for (unsigned int x = x0; x < x1; x++)
	{
		for (unsigned int y = y0; y < y1; y++)
		{
			GLfloat * pos =  &vertexbuffer[(y*s+x)*3];
			pos[0] = x;
			pos[1] = h[y][x];
			pos[2] = y;
		}
	}
	

	for (unsigned int x = x0; x < x1; x++)
	{
		for (unsigned int y = y0; y < y1; y++)
		{
			GLfloat * pos =  &normalbuffer[(y*s+x)*3];
			pos[0] = vns[y][x](0);
			pos[1] = vns[y][x](1);
			pos[2] = vns[y][x](2);
		}
	}


#ifdef HEAVY_DUTY_CHECKS
	for (unsigned int x = 0; x < s; x++)
	{
		for (unsigned int y = 0; y < s; y++)
		{
			GLfloat * pos =  &vertexbuffer[(y*s+x)*3];
			if	(	pos[0] == x &&
					abs(pos[1] - h[y][x]) < 0.00001 &&
					pos[2] == y )
			{
			}
			else
			{
				cout<< "outlier: (" << x << ", " << y << ")  is "
					<< pos[0] << " ? " << x << " / " 
					<< pos[1] << " ? " << h[y][x] << " / "
					<< pos[2] << " ? " << y << "  in rect "
					<< "(" << x0 << ", " << y0 << ") - "
					<< "(" << x1 << ", " << y1 << ")" << " but is not updated" << endl;
				assert(false);
			}
		}
	}


	for (unsigned int x = 0; x < s; x++)
	{
		for (unsigned int y = 0; y < s; y++)
		{
			GLfloat * pos =  &normalbuffer[(y*s+x)*3];
			if	(	abs(pos[0] - 	vns[y][x](0)) < 0.000001 &&
					abs(pos[1] - 	vns[y][x](1)) < 0.000001 &&
					abs(pos[2] - 	vns[y][x](2)) < 0.000001)
			{
			}
			else
			{
				cout<< "outlier: (" << x << ", " << y << ")  is "
					<< pos[0] << " ? " << vns[y][x](0) << " / "
					<< pos[1] << " ? " << vns[y][x](1) << " / "
					<< pos[2] << " ? " << vns[y][x](2) << "  in rect "
					<< "(" << x0 << ", " << y0 << ") - "
					<< "(" << x1 << ", " << y1 << ")" << " but is not updated" << endl;

				assert(false);	
			}
		}
	}
#endif // HEAVY_DUTY_CHECKS	

}


void GridModel::subtractSphere(const Vector<GLfloat> & p, double radius)
{
	assert(*p == 3);
	
	double x = p(0) + s/2.0;
	double y = p(2) + s/2.0;
	double z = p(1);

	int x0 = (int)ceil(x-radius); if (x0 < 0) 		x0 = 0;
	int y0 = (int)ceil(y-radius); if (y0 < 0) 		y0 = 0;
	int x1 = (int)ceil(x+radius); if (x1 > (int)s) 	x1 = s;
	int y1 = (int)ceil(y+radius); if (y1 > (int)s) 	y1 = s;

	// totally out...
	if (x0 >= (int)s || x1 <= 0 ||
		y0 >= (int)s || y1 <= 0) return;

	// inconsistent
	if (x0 >= x1 || y0 >= y1) return; // no change
	
	for (int xi = x0; xi < x1; xi++)
	for (int yi = y0; yi < y1; yi++)
	{
			// also add some 'shake'
			h[yi][xi] = radius*mod(0.1) + min(h[yi][xi], z-sqrt(sqr(radius)-sqr(x-xi)-sqr(y-yi)));
	}

	// widen 1 before smoothing for better cavein simulation
	x0 = w_max(x0-1, 0); x1 = w_min(x1+1, (int)s);
	y0 = w_max(y0-1, 0); y1 = w_min(y1+1, (int)s);
	smooth(x0, x1, y0, y1);
		
	// because of neighbour calculation for normals, we need to widen 1 for the changes
	x0 = w_max(x0-1, 0); x1 = w_min(x1+1, (int)s);
	y0 = w_max(y0-1, 0); y1 = w_min(y1+1, (int)s);

	// recompute changed geometry
	computeFaceNormals  	(x0, x1, y0, y1);
	computeVertexNormals  	(x0, x1, y0, y1);
	bufferize 	(x0, x1, y0, y1);				

	// invalidate concerned display lists
	//    x0-1 is the first affected vertex/normal, x1 the last

    unsigned int pgx = s/patchesx;
    unsigned int pgy = s/patchesy;

    // divise and round down to get patch from point
    int px0 = w_max(0, (x0-1)/pgx);
    int py0 = w_max(0, (y0-1)/pgy);
 	// to use our former convention, p<x>1 is one too high
    int px1 = w_min(patchesx, x1/pgx+1);
    int py1 = w_min(patchesy, y1/pgy+1);

	
	for (int xi = px0; xi < px1; xi++)
	for (int yi = py0; yi < py1; yi++)
		patches[xi][yi]->invalidate();
	

#ifdef HEAVY_DUTY_CHECKS	
	#define out(X0, X1, Y0, Y1)	((int)(X1) < (int)(Y0) || (int)(X0) > (int)(Y1))
	// check
	for (unsigned int xi = 0; xi < patchesx; xi++)
	for (unsigned int yi = 0; yi < patchesy; yi++)
	{
		if( out(patches[xi][yi]->x0, patches[xi][yi]->x1, x0, x1) ||
			out(patches[xi][yi]->y0, patches[xi][yi]->y1, y0, y1) ||
			patches[xi][yi]->bound == NULL)
		{
		}
		else
		{
			cout    << "outlier: [" << xi << "][" << yi << "]  "
					<< "(" << patches[xi][yi]->x0 << ", " << patches[xi][yi]->y0 << ") - "
					<< "(" << patches[xi][yi]->x1 << ", " << patches[xi][yi]->y1 << ") lies in demolition zone "
					<< "(" << x0 << ", " << y0 << ") - "
					<< "(" << x1 << ", " << y1 << ")" << " but is not updated" << endl;
			assert(false);					
		}
	}
#endif // HEAVY_DUTY_CHECKS
	
	
}
