#include "WaterModel.h"

#include <cstdlib>
#include "utils.h"
#include "array.h"
#include "GL/Material.h"
#include "GL/HomLinAlg.h"


Material WaterModel::WaterMaterial = Material(
 Vector<GLfloat>(0.3, 0.3, 1.0, 0.5),
 Vector<GLfloat>(0.3, 0.3, 1.0, 0.5),
 Vector<GLfloat>(0.4, 0.4, 1.0, 0.5),
 Vector<GLfloat>(0.0, 0.0, 0.0, 0.0),
 0.3);



WaterModel::WaterModel(double size, unsigned int n, unsigned int frames) 
	: size(size), n(n), frames(frames)
{
	vertexbuffer = 	new GLfloat[3 * n * n];
	normalbuffer = 	new_array<GLfloat>(frames, 3 * n * n);
	indexbuffer = 	new_array<GLuint>(n-1, 2*n);

	TIME( bufferize() )
}


WaterModel::~WaterModel()
{
	delete [] vertexbuffer;
	delete_array(normalbuffer, frames, 	3*n*n);
	delete_array(indexbuffer, 	n-1, 2*n);
}



void WaterModel::drawVertexNormals(double length, int currentFrame) const
{
	OpenGL::Disable(GL_LIGHTING);

	OpenGL::Color(1, 0, 1);
	OpenGL::Begin(GL_LINES);

	// for every vertex
	for (unsigned int x = 0; x < n; x++)
	{
		for (unsigned int y = 0; y < n; y++)
		{
			Vector<GLfloat> v(x, 0, y);
			Vector<GLfloat> n(	normalbuffer[currentFrame][(y*n+x)*3+0],
								normalbuffer[currentFrame][(y*n+x)*3+1],
								normalbuffer[currentFrame][(y*n+x)*3+2]);
			OpenGL::Vertex(v);
			OpenGL::Vertex(v + n * length);
		}
	}
	OpenGL::End();				
	OpenGL::Enable(GL_LIGHTING);					
}




void WaterModel::draw(int currentFrame, double waterlevel) const
{
	double d = n/-2.0;   // centering displacement
	
	WaterMaterial.set();
	
	OpenGL::PushMatrix();
	OpenGL::MultMatrix(HomMatrix<GLfloat>::translation(Vector<GLfloat>(0,waterlevel,0)));	
	OpenGL::MultMatrix(HomMatrix<GLfloat>::scaling(4, size/n));
	OpenGL::MultMatrix(HomMatrix<GLfloat>::translation(Vector<GLfloat>(d,0,d)));
	

	OpenGL::VertexPointer(3, GL_FLOAT, 0, vertexbuffer);
	OpenGL::NormalPointer(   GL_FLOAT, 0, normalbuffer[currentFrame]);
	OpenGL::EnableClientState(GL_VERTEX_ARRAY);
	OpenGL::EnableClientState(GL_NORMAL_ARRAY);
	
	
	for (unsigned int x = 0; x < n-1; x++)
	{
		OpenGL::DrawElements(GL_TRIANGLE_STRIP, n*2, GL_UNSIGNED_INT, indexbuffer[x]);
	}
	
//	drawVertexNormals(1, currentFrame);
	
	OpenGL::PopMatrix();

	OpenGL::DisableClientState(GL_VERTEX_ARRAY);
	OpenGL::DisableClientState(GL_NORMAL_ARRAY);
}



void WaterModel::bufferize()
{
	
//	int d = (signed)s/2;

	for (unsigned int x = 0; x < n; x++)
	{
		for (unsigned int y = 0; y < n; y++)
		{
			GLfloat * pos =  &vertexbuffer[(y*n+x)*3];
			pos[0] = x;
			pos[1] = 0;
			pos[2] = y;
		}
	}

	for (unsigned int x = 0; x < n; x++)
	{
		for (unsigned int y = 0; y < n; y++)
		{
			double nx = ((double)rand()/RAND_MAX-0.5)*2;
			double nz = ((double)rand()/RAND_MAX-0.5)*2;
			
			double df = (double)rand()/RAND_MAX;
			
			for (unsigned int f = 0; f < frames; f++)
			{
				GLfloat * pos =  &normalbuffer[f][(y*n+x)*3];
				double s = sin(2*M_PI*((double)f/frames + df));
				
				Vector<GLfloat> N = Vector<GLfloat>(nx * s, 1, nz *s).normalize();
				
				pos[0] = N(0);
				pos[1] = N(1);
				pos[2] = N(2);
			}
		}
	}
	
	for (unsigned int x = 0; x < n-1; x++)
	{
		GLuint * pos = indexbuffer[x];
		
		for (unsigned int y = 0; y < n; y++)
		{
			for (unsigned int u = 0; u < 2; u++)
			{
				*pos++ = x+1-u + y*n;
			}
		}
	}
}
